     H Option(*SRCSTMT:*NODEBUGIO)
     H Debug(*YES)
     H DatFmt(*ISO)
     H TimFmt(*ISO)
     H FixNbr(*Zoned: *InputPacked)

       // TNAPI 3.0
       // API for sending and receiving data via telnet

       //  TNAPI is shareware. You are free to use it as long as you don't
       //   represent it as your own work

       //   Copyright 2017 Albert York

       //  To compile: CRTBNDRPG DFTACTGRP(*NO) BDNDIR(QC2LE)

       // ---------------------------------------------------------------

       //  Incoming parameters:
       //  - Send buffer (256) - Data to send to the iSeries (can
       //     include imbedded cursor control codes).
       //
       //  - Function key (6) - Function key to send to the iSeries
       //     or other special functions.
       //
       //  - Timeout (5.1) Maximum number of seconds to wait for a
       //     response from the iSeries
       //
       //  - Wait string array (740)
       //     Divided into 20X37 elements as follows
       //       Pos 1-1  - And/Or option. A=and, O=or. First element or
       //                  blank defaults to O
       //       Pos 2-3  - Row to wait for text
       //       Pos 4-5  - Column to wait for text
       //       Pos 6-37 - Text string to wait for
       //                  If there are trailing blanks then append
       //                  %CR
       //     If no wait string then pass a blank field of 740 bytes.
       //
       //  - System (64). Can be a system name or IP address.
       //     If blank then LOOPBACK is assumed.
       //
       //  - Display screens (Y/N). Display screens as they are received
       //    (interactive jobs only)
       //
       //  - Return screen attributes (Y/N) - Return attributes as well
       //     as data.
       //
       //  - Log file name. Specify *NONE if no logging. If blank
       //     TNAPILOG is assumed. other values:
       //     *JOBNAME - A mbr in TNAPILOG is created using the job name
       //     *USER - A mbr in TNAPILOG is created using the user ID
       //     *JOBNBR - A mbr in TNAPILOG is created using the job no
       //      with the users name is created. If *JOBNBR is specified
       //      a member with the job number is created.
       //     Log files are flat files 80 bytes in length and
       //       logs the data that is sent and each screen that is
       //       received
       //     TNAPIBUFF - a flat file of 8030 characters. if it exists
       //       the raw data being sent to and from the iSeries is
       //       logged to it

       // ---------------------------------------------------------------

       //  Returned parameters:
       //  - Wait string pointer (2.0) - Pointer to the array element
       //     of the first starting wait sequence which matched
       //
       //  - Return code (1) - 0 = Normal return
       //                      1 = Error on screen requiring ERROR RESET
       //                      2 = Timeout waiting for the iSeries
       //                      3 = Socket error or EOJ
       //
       //  - Screen data (1920) - Current screen characters (24 X 80)
       //
       //  - Screen attributes (1920) - Display attributes
       //
       //  - Row - (2.0) - Current row Position
       //
       //  - Column - (2.0) - Current column Position

       // ---------------------------------------------------------------

       //  Notes:
       //    If wait strings are specified the program will wait until the
       //     strings appear at the specified row and column, or until the
       //     specified timeout has elapsed. Wait strings can be a
       //     combination of AND/OR lines.

       //    If no wait string is specified and a function key is specified
       //     the program will wait until the iSeries has completed sending
       //     the screen or the specified timeout has elapsed.

       //    If neither is specified the program will send the data and
       //     retrieve the current screen but will not wait.

       //    Valid function keys are ENTER, F1-F24, PRINT, etc. See the
       //      list at the end of the program. Must be upper case.
       //     Special functions:
       //      POSCSR - position cursor. Desired cursor position
       //               is in PARMSENDBUFF as RR,CC (ex 03,27)
       //      REDRAW - Send request to iSeries to resend the current screen
       //      GETDTA - Do not send anything. Just receive data and return
       //      EOJ    - Close socket, set on LR and return

       //    To include cursor control codes in the send buffer prefix
       //     them with % (ie %TB). All codes are two uppercase characters.
       //     See the table at the end of this program for valid codes
       //      Special control codes:
       //       %L0 - do not log data that comes after %L0
       //              until %L1 or end of data
       //              (used primarily for passwords)
       //       %L1 - Resume logging data
       //       %PC - Send per cent symbol (%)
       //       %CR - End of data. Optional (used for trailing blanks)

       //    The returned attributes follow the iSeries standard (ie x'20'
       //       is a normal display). Color codes are not supported.

       //    The file descriptor routines were provided by Scott Klement
       //      and are marked with 'SK'

      // Log file (optional flat file of 80 bytes)
     FTNAPILOG  o    f   80        disk    Usropn
     FTNAPIBUFF o    f 8030        disk    Usropn

      // Display file
     FTNAPIDSPF cf   e             workstn usropn

     D AttrStr         s              4
     D Bufflen         s             10i 0
     D c               s             10i 0
     D CharPtr         s             10I 0
     D CheckPtr        s             10i 0
     D Cmd             s            256
     D CmdError        s               n
     D CmdLen          s             15  5 inz(256)
     D CurrCol         s              2  0
     D CurrAttr        s              1
     D CurrRow         s              2  0
     D CsrLocStr       s            400
     D DataWaiting     s               n
     D DateTime        s             23
     D ESCSeq          s               n   inz(*off)
     D DisplayScreen   s               n
     D FirstNum        s              2  0
     D FirstTime       s               n   inz(*on)
     D FromPtr         s             10i 0
     D JobMbr          s             10
     D L0              s             10i 0
     D L1              s             10i 0
     D LastSystem      s             64
     D LogBuff         s            256
     D LogBuffer       s               n
     D LogRow          s              2  0
     D LogCol          s              2  0
     D LogDta          s             80
     D LogField        s             50
     D Logging         s               n   inz(*off)
     D LogJob          s             28
     D LogSeperator    s             80    inz(*ALL'=')
     D LogSystem       s             10
     D MicroSec        s             15p 0
     D MoveToBtm       s              5
     D MoveToTop       s              4
     D NormalDisplay   s              4
     D p               s             10i 0
     D PosCol          s              2  0
     D PosCount        s              2  0
     D PosCtl          s              1
     D PosRow          s              2  0
     D r               s             10i 0
     D ReadSet         s             28
     D RecvCnt         S             10i 0
     D ReadBuff        s           8000
     D ReadCnt         s             10i 0
     D ReadLim         s             10i 0 inz(%len(ReadBuff))
     D RecvEOT         s              8
     D result          s             10i 0
     D rc              S             10i 0 INZ(0)
     D s               s             10i 0
     D SaveSet         s             28
     D SearchFct       s              2
     D SendLen         S             10i 0 inz(0)
     D SendEOT         s              6
     D Seperator       s             80    inz(*ALL'#')
     D TAttr           s              1
     D Time1           s               z
     D Time2           s               z
     D Timeout         s             15P 0
     D UnderlineRcvd   s               n
     D WaitForData     s               n   inz(*off)
     D SocketID        S             10i 0 INZ(-1)
     D SVRADDRLEN      S             10u 0
     D PosCsrStr       s              2
     D Tchar           s              1
     D w1              s             10i 0
     D w2              s             10i 0
     D w3              s             10i 0
     D WaitCnt         s             10i 0
     D WaitOpt         s              1    dim(20)
     D WaitPos         s              4  0 dim(20)
     D WaitKey         s             32    dim(20)
     D WaitLen         s              3  0 dim(20)
     D WaitCompStr     s             32
     D x               s             10i 0

      * Constants
     D AF_INET         C                   CONST(2)
     D BellChar        c                   const(x'2F')
     D ESCChar         c                   const(x'27')
     D FIONBIO         c                   CONST(131074)
     D INADDR_NONE     C                   CONST(4294967295)
     D LeftBracket     c                   const(x'4A')
     D NONBLOCKIO      s             10i 0 inz(1)
     D NormalAttr      c                   const(x'20')
     D NullChar        c                   const(x'00')
     D OneSecond       c                   const(1000000)
     D PositionCursor  c                   const('POSCSR')
     D QuarterSecond   c                   const(250000)
     D SOCK_STREAM     C                   CONST(1)
     D TenthSecond     c                   const(100000)
     D ThirtySeconds   c                   const(30000000)
     D TwoMinutes      c                   const(120000000)
     D UNUSED          s             10i 0 INZ(0)

     D ScreenChar      ds          1920
     D   ScreenRow                   80    dim(24)
     D   ScreenCol             1   1920    dim(1920)

     D ScreenAttr      ds          1920
     D   AttrCol                      1    dim(1920)

     D LogScreenBuff   ds            80
     D  LogScreenChar                 1    dim(80)

     D SendBuff        ds          1024
     D  SendChar                      1    dim(1024)

     D RecvBuff        ds          6000
     D  RecvChar                      1    dim(6000)

     D AttrControl     ds
     D  White                  1      1
     D  Underline              2      2
     D  ReverseImage           3      3
     D  Blink                  4      4

     D ParmWaitArr     ds
     D  WaitArr                      37    dim(20)

     D WaitCtl         ds
     D  WaitOption             1      1
     D  WaitRow                2      3  0
     D  WaitCol                4      5  0
     D  WaitStr                6     37

     D PosString       ds
     D  PosRow$                1      2
     D  PosCol$                4      5

     D                 ds
     D CharVal                 1      2
     D  CharVal1               1      1
     D  CharVal2               2      2
     D  NumVal                 3      4s 0
     D  NumVal2                4      4s 0

     D TimeStamp       ds
     D  CurrTime                       Z

       //  Function key conversion table
     D FKName          S              6    DIM(43) CTDATA PERRCD(1)
     D FKSeq           S              5    DIM(43) CTDATA PERRCD(1)

       //  Cursor control conversion table
     D CCName          S              2    DIM(15) CTDATA PERRCD(1)
     D CCSeq           S              5    DIM(15) CTDATA PERRCD(1)

     D ParmScreenChar  ds          1920
     D   ScrnData              2   1920

     D SeqField        ds
     D  SeqNum1                1      2s 0
     D  SeqChar1               3      3
     D  SeqNum2                3      4s 0
     D  SeqChar2               5      5
     D  SeqNum3                3      5s 0

     D                 ds
     D  BinaryValue            1      2b 0 inz(0)
     D  Char                   2      2

     D SvrAddress      DS
     D   Svrfamily                    5i 0
     D   Svrport                      5u 0
     D   Svraddr                     10u 0
     D   SvrNull                      8a

     D p_hostent       S               *

     D hostent         DS                  Based(p_hostent)
     D   h_name                        *
     D   h_aliases                     *
     D   h_addrtype                  10I 0
     D   h_length                    10I 0
     D   h_addr_list                   *

     D p_h_addr        S               *   Based(h_addr_list)
     D h_addr          S             10U 0 Based(p_h_addr)

      // Clear parts of the screen
     D ClearData       PR
     D   StartRow                     2  0 const
     D   StartCol                     2  0 const
     D   EndRow                       2  0 const
     D   EndCol                       2  0 const

      // Move character to buffer
     D MoveCharToBuff  PR

      // EBCDIC/ASCII translation
     D Translate       PR                  ExtPgm('QDCXLATE')
     D                                5P 0 CONST
     D                             1024A   options(*varsize)
     D                               10A   CONST
     D                               10A   CONST

      // Run commands
     D ExecCmd         pr                  extpgm('QCMDEXC')
     D                              256
     D                               15  5

      // Socket prototypes
     D GetHostId       PR            10u 0 ExtProc('gethostid')

     D GetHostByName   PR              *   extproc('gethostbyname')
     D                                 *   value options(*string)

     D inet_addr       PR            10U 0 ExtProc('inet_addr')
     D                                 *   Value Options(*String)

     D Socket          PR            10i 0 ExtProc('socket')
     D                               10i 0 Value
     D                               10i 0 Value
     D                               10i 0 Value

     D Read            PR            10i 0 ExtProc('read')
     D                               10i 0 Value
     D                                 *   Value
     D                               10i 0 Value

     D Write           PR            10i 0 ExtProc('write')
     D                               10i 0 Value
     D                                 *   Value
     D                               10i 0 Value

     D Connect         PR            10i 0 ExtProc('connect')
     D                               10i 0 Value
     D                                 *   Value
     D                               10u 0 Value

     D Close           PR            10i 0 ExtProc('close')
     D                               10i 0 Value

     D ioctl           PR            10i 0 ExtProc('ioctl')
     D                               10i 0 value
     D                               10i 0 value
     D                               10i 0

     D timeval         ds
     D  tv_sec                       10i 0
     D  tv_usec                      10i 0

      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      * SK
      * Set a File Descriptor in a set ON...  for use w/Select()
      *
      *      peFD = descriptor to set on
      *      peFDSet = descriptor set
      *
     D FD_SET          PR
     D   peFD                        10I 0
     D   peFDSet                     28A

      *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

     D CheckForData    pr              n
     D  SocketID                     10i 0
     D  TimeOut                      15p 0

     D                sds
     D SDSJobName            244    253
     D SDSUser               254    263
     D SDSJob#               264    269

      *  Parameters
     C     *entry        plist
     C                   parm                    ParmSendbuff    256
     C                   parm                    ParmFctKey        6
     C                   parm                    ParmWaitTime      5 1
     C                   parm                    ParmWaitArr
     C                   parm                    ParmSystem       64
     C                   parm                    ParmDisplay       1
     C                   parm                    ParmAttr          1
     C                   parm                    ParmLogFile      10
     C                   parm                    ParmWaitPtr       2 0
     C                   parm                    ParmReturn        1
     C                   parm                    ParmScreenChar
     C                   parm                    ParmScreenAttr 1920
     C                   parm                    ParmCurrRow       2 0
     C                   parm                    ParmCurrcol       2 0

      /free

       //  End of job
       if ParmFctKey = 'EOJ';
         exsr Eoj;
       endif;

       // Close socket and return
       if ParmFctKey = 'CLOSE';
         rc = close(socketid);
         socketid = -1;
         FirstTime = *on;
         return;
       endif;

       // Test connection
       if ParmFctKey = 'TSTCON';
         SendBuff = ESCChar + LeftBracket + 'C'
          + ESCChar + LeftBracket + 'D';
         SendLen = %len(%trim(SendBuff));
         exsr WriteData;
         return;
       endif;

       //  Default to normal return
       ParmReturn = '0';

       // Write string to the log file
       if ParmFctKey = 'LOGDTA';
         if Logging;
           LogDta = ParmSendBuff;
           CurrTime = %timestamp();
           DateTime = %subst(TimeStamp:1);
           *in26 = *on;
           except;
           *in26 = *off;
         endif;
         return;
       endif;

       //  Start the socket connection the first time called
       //   or if the system changes
       if FirstTime
          or ParmSystem <> LastSystem;
         exsr ConnectToiSeries;
       endif;

       //  Get any transmissions that haven't been received yet
       TimeOut = *zeros;
       exsr RecvData;

       //  Just retrieve data and return
       if ParmFctKey = 'GETDTA';
         if Logging;
           exsr LogFctKey;
         endif;
       else;

       //  Send data to iSeries
         if ParmSendBuff <> *blanks;
           exsr SendBuffer;
         endif;

         WaitForData = *off;

       //  Parse wait string parameters
         exsr ParseWaitStr;

       //  Send function key
         if ParmFctKey <> *blanks;

       //  Log function key
           if Logging;
             exsr LogFctKey;
           endif;

       //  Do not send position cursor command
           if ParmFctKey <> PositionCursor;

       //  Look up function key in conversion table
             p = %lookup(ParmFctKey:FKName);

       //  Valid function key. Convert to ASCII sequence.
             if p > 0;
               WaitForData = *on;
               SeqField = FKSeq(p);
               BinaryValue = SeqNum1;
               SendBuff = char;
               if SeqChar1 > *blank;
                 BinaryValue = SeqNum3;
                 SendBuff = %trim(SendBuff) + char;
               endif;
       //  Add end of transmission marker if no wait string
               if WaitCnt = 0;
                 SendBuff = %trimr(SendBuff) + SendEOT;
               endif;
               exsr WriteData;
             endif;
           endif;
         endif;

       //  Wait for a response if a function key was sent or a compare
       //   string was specified.
         if WaitForData = *on;
           Timeout = (ParmWaitTime*OneSecond);
           exsr RecvData;
         endif;

       endif;

       //  Return screen attributes
       if ParmAttr = 'Y';
         ScreenAttr = %xlate(' ':NormalAttr:ScreenAttr);
         ParmScreenAttr = ScreenAttr;
       endif;

       //  Format return parameters
       ParmScreenChar = ScreenChar;
       ParmCurrRow = CurrRow;
       ParmCurrcol = CurrCol;

       // Log current screen
       if Logging;
         exsr LogScreen;
       endif;

       // Display current screen
       if DisplayScreen;
         write TNAPISCRN;
       endif;

       return;

       // *********************************************************************

       //  Move text to send buffer and
       //   convert cursor control codes to ASCII sequences

       begsr SendBuffer;

         BuffLen = %len(%trimr(ParmSendBuff));

       //  Log data that is being sent
         if Logging;

           s = 1;
           *in21 = *on;
           LogBuff = ParmSendBuff;

       //  Check for logging commands L0/L1
           if BuffLen > 3;
             dow '1';
               L0  = %scan('%L0':LogBuff:s) + 3;
               if L0 < 4;
                 leave;
               endif;
               L1 = %scan('%L1':LogBuff:L0);
               if L1 = 0;
                 L1 = BuffLen+1;
               endif;
               p = L1-L0+1;
               %subst(LogBuff:L0:p) = *all'*';
               s = L1 + 3;
               if s > BuffLen;
                 leave;
               endif;
             enddo;
           endif;

           s = 1;

       //  Log data in 50 byte chunks
           dow '1';
             if s > Bufflen;
               leave;
             endif;
             LogField = %subst(LogBuff:s);
             if %subst(LogField:1:1) = NullChar;
               leave;
             endif;
             *in21 = *on;
             exsr Logdata;
             s = s + 50;
           enddo;

         endif;

         select;

       //  Position cursor
         when ParmFctKey = PositionCursor;
           PosString = ParmSendBuff;
           exsr MoveCursor;
           SendBuff = CsrLocStr;
           Sendlen = %len(%trim(SendBuff));

         other;
       //  Process send buffer one character at a time
           for FromPtr = 1 to BuffLen;

             Char = %subst(ParmSendBuff:Fromptr:1);

       //  Null can be used for trailing blanks
             if char = NullChar;
               iter;
             endif;

       //  Regular character. Move it to send buffer
             if char <> '%';
               MoveCharToBuff();
               iter;
             endif;

       //  % indicates start of cursor control code
             SearchFct = %subst(ParmSendBuff:FromPtr+1:2);

       //  %L0/L1 are logging commands - just ignore them
             if SearchFct = 'L0' or SearchFct = 'L1';
               FromPtr = FromPtr + 2;
               iter;
             endif;

       //  %% means send per cent symbol
             if SearchFct = '%%';
               char = '%';
               MoveCharToBuff();
               iter;
             endif;

       //  %PC means position cursor
             if SearchFct = 'PC';
               FromPtr = FromPtr + 3;
               PosString = %subst(ParmSendBuff:FromPtr:5);
               exsr MoveCursor;
               SendBuff = %trim(SendBuff) + CsrLocStr;
               Sendlen = %len(%trim(SendBuff));
       // Fromptr will be incremented by the FOR loop
               FromPtr = FromPtr + 4;
               iter;
             endif;

       //  %CR means end of line (if trailing blanks)
             if SearchFct = 'CR';
               leave;
             endif;

       //  Look up cursor control code in conversion table
             p = %lookup(SearchFct:CCName);

       //  Not a valid code - must be just %
             if p = 0;
               char = '%';
               MoveCharToBuff();
               iter;
             endif;

       //  Convert code to ASCII sequence
             SeqField = CCSeq(p);
             BinaryValue = SeqNum1;
             MoveCharToBuff();
             if SeqNum2 <> 0;
               BinaryValue = SeqNum2;
               MoveCharToBuff();
             endif;
             if SeqChar2 <> *blanks;
               char = SeqChar2;
               MoveCharToBuff();
             endif;
             FromPtr = FromPtr + 2;

           endfor;

         endsl;

         WaitCnt = 0;
         ParmWaitPtr = 0;

       //  Add end of transmission sequence
         %Subst(SendBuff:SendLen+1:6) = SendEOT;
         SendLen = SendLen + 6;

       //  Send data to the iSeries
         exsr WriteData;

       //  Wait for text to be echoed
         Timeout = (ParmWaitTime*OneSecond);
         exsr RecvData;

       endsr;

       // *********************************************************************

       //  Write data to the socket
       begsr WriteData;

       //  Find the length if not specified
         if SendLen = 0;
           SendLen = %len(%trim(SendBuff));
         endif;

         if LogBuffer;
           *in28 = *on;
           CurrTime = %timestamp();
           DateTime = %subst(TimeStamp:1);
           except;
           *in28 = *off;
         endif;

       //  Convert to ASCII
         Translate(SendLen:SendBuff:'QASCII':'QSYS');

       //  Write to the socket
         rc = Write(SocketID:%addr(SendBuff):SendLen);

       // Socket error
         if rc < 0;
           if Logging;
             LogDta = 'Socket error';
             CurrTime = %timestamp();
             DateTime = %subst(TimeStamp:1);
             *in26 = *on;
             except;
             *in26 = *off;
           endif;
           exsr EOJ;
         endif;

         clear SendBuff;
         SendLen = *zeros;

       endsr;

       // *********************************************************************

       //  Receive data and place in screen buffer

       begsr RecvData;

         RecvCnt = *zeros;
         time1 = %timestamp();

         exsr ReadData;

         if RecvCnt > 0;
           exsr ParseData;
         endif;

       endsr;

       // *********************************************************************

       begsr ReadData;

         tv_sec = 0;
         tv_usec = TimeOut;
         time1 = %timestamp();

         dow '1';

           dow '1';

             DataWaiting = CheckForData(SocketID:TimeOut);

             if not DataWaiting;
               leave;
             endif;

             clear ReadBuff;

       //  Read data from the socket
             ReadCnt =  Read(SocketID:%addr(ReadBuff):ReadLim);

       //  No data in the buffer
             if ReadCnt = 0;
               leave;
             endif;

       //  Have received data

       //  Default to one tenth second wait after initial read
             if TimeOut > 0;
               tv_Usec = TenthSecond;
             endif;

       //  Convert to EBCDIC
             Translate(ReadCnt:ReadBuff:'QEBCDIC':'QSYS');
             if LogBuffer;
               *in27 = *on;
               CurrTime = %timestamp();
               DateTime = %subst(TimeStamp:1);
               except;
               *in27 = *off;
             endif;

       //  Add to receive buffer
             %subst(RecvBuff:RecvCnt+1:ReadCnt) = ReadBuff;
             RecvCnt = RecvCnt + ReadCnt;

       //  Sometime 'ESC+[+0m' (set to normal display) is stuck on the
       //  end. Just ignore it
             CheckPtr = RecvCnt;
             if CheckPtr > 3 and
                %subst(RecvBuff:Checkptr-3:4) = NormalDisplay;
               Checkptr = CheckPtr - 4;
             endif;

       //  Check for end of transmission strings
             if CheckPtr > 7 and
                %subst(RecvBuff:CheckPtr-7:8) = RecvEOT;
               leave;
             endif;

       //  End of transmission string will come back as row and column
       //   positioning commands if the cursor is on row 24
             if Checkptr> 12
                and  %subst(RecvBuff:CheckPtr-5:4) = MovetoTop
                and %subst(RecvBuff:CheckPtr-12:5)= MoveToBtm;
               leave;
             endif;
             if Checkptr > 14
                and %subst(RecvBuff:CheckPtr-6:4) = MovetoTop
                and %subst(RecvBuff:CheckPtr-14:5)= MoveToBtm;
               leave;
             endif;

           enddo;

       //  Not waiting for text to appear on the screen
           if WaitCnt = 0;
             leave;
           endif;

       //  Check data on the screen
           exsr ParseData;
           exsr CheckWaitStr;
       //  Text has appeared
           if ParmWaitPtr > 0;
       //  Send EOT string to get the rest of the screen
             WaitCnt = 0;
             SendBuff = SendEOT;
             exsr WriteData;
             leave;
           endif;
        // See if timeout has been surpassed
           time2 = %timestamp();
           MicroSec = %diff(time2:time1:*Mseconds);
           if MicroSec > Timeout;
             ParmReturn = '2';
             *in24 = *on;
             leave;
           endif;
        // Reduce timeout by elapsed time
           TV_usec = Timeout - MicroSec;

         enddo;

       endsr;

       // *********************************************************************

       //  Parse data and format screen
       begsr ParseData;

         for CharPtr = 1 to RecvCnt;

           Char = RecvChar(CharPtr);

       //  Possible error
           if char = BellChar;
             Parmreturn = '1';
             iter;
           endif;

       //  ESC indicates beginning of formatting sequence
           if Char = ESCChar;
             ESCSeq = *on;
             clear CharVal;
             clear AttrStr;
             iter;
           endif;

       //  If not in an ESC sequence place data in screen buffer
           if ESCSeq <> *on;
             if Char >= *blank;
               p = ((CurrRow-1)*80) + Currcol;
               ScreenCol(p) = Char;
               AttrCol(p) = CurrAttr;
       //  Move cursor to the right
               NumVal = 1;
               exsr CursorRight;
             endif;
             iter;
           endif;

       //  Processing an escape sequence

       //  Ignore certain characters in ESC sequence
           if Char < *blank or Char = '?' or Char = LeftBracket;
             iter;
           endif;

       //  Save numeric values
           if Char >='0' and Char <= '9';
             CharVal = %trimr(CharVal) + Char;
             AttrStr = %trimr(AttrStr) + Char;
             iter;
           endif;

       //  Convert numeric values to number
           NumVal = *zeros;
           if CharVal1 <> *blanks;
             if CharVal2 = *blanks;
               NumVal2 = %int(CharVal1);
             else;
               NumVal = %int(CharVal);
             endif;
           endif;

       //  Parameter seperator
           if Char = ';';
             FirstNum = NumVal;
             clear CharVal;
             iter;
           endif;

       //  End of ESC sequence
           ESCSeq = *off;

           select;

       //  Up n lines
           when Char = 'A';
             exsr CursorUp;

       //  Down n lines
           when Char = 'B';
             exsr CursorDown;

       //  Right n Positions
           when Char = 'C';
             exsr CursorRight;

       //  Left n Positions
           when Char = 'D';
             exsr CursorLeft;

       //  Position to row and column
           when Char = 'H'
              or Char = 'r';
             CurrRow = FirstNum;
             CurrCol = NumVal;

       //  Clear parts of the screen
           when Char = 'J';
             exsr ClearScreen;

       //  Clear parts of the current row
           when Char = 'K';
             exsr ClearRow;

       //  Set display attribute
           when Char = 'm';
             exsr SetAttr;

           endsl;

         endfor;

         RecvCnt = *zeros;

       endsr;

       // *********************************************************************

       begsr CursorUp;

       //  move cursor up x rows
         if NumVal = 0;
           Numval = 1;
         endif;

         PosRow = NumVal;

         for r = 1 to PosRow;
           CurrRow = CurrRow - 1;
           if CurrRow < 1;
             CurrRow = 24;
           endif;
         endfor;

       endsr;

       // *********************************************************************

       begsr CursorDown;

       //  move cursor down x rows
         if NumVal = 0;
           Numval = 1;
         endif;

         PosRow = NumVal;

         for r = 1 to PosRow;
           CurrRow = CurrRow + 1;
           if CurrRow > 24;
             CurrRow = 1;
           endif;
         endfor;

       endsr;

       // *********************************************************************

       begsr CursorRight;

       //  move cursor right x columns
         if NumVal = 0;
           Numval = 1;
         endif;

         PosCol = numVal;

         for c = 1 to PosCol;
           CurrCol = CurrCol + 1;
           if CurrCol > 80;
             CurrCol = 1;
             NumVal = 1;
             exsr CursorDown;
           endif;
         endfor;

       endsr;
       // *********************************************************************

       begsr CursorLeft;

       //  move cursor left x columns
         if NumVal = 0;
           Numval = 1;
         endif;

         PosCol = NumVal;

         for c = 1 to PosCol;
           CurrCol = CurrCol - 1;
           if CurrCol < 1;
             CurrCol = 80;
             Numval = 1;
             exsr CursorUp;
           endif;
         endfor;

       endsr;

       // *********************************************************************

       begsr ClearScreen;

         select;

       //  Clear from current Position to end of screen
         when NumVal = 0;
           ClearData(CurrRow:CurrCol:24:80);

       //  Clear from start of screen to the current position
         when NumVal = 1;
           ClearData(1:1:CurrRow:CurrCol);

       //  Clear the entire screen
         when NumVal = 2;
           clear ScreenChar;
           clear ScreenAttr;
           UnderlineRcvd = *off;

         endsl;

       endsr;

       // *********************************************************************

       begsr ClearRow;

         select;

       //  Clear from the current position to the end of the row
         when NumVal = 0;
           ClearData(CurrRow:CurrCol:CurrRow:80);

       //  Clear from the start of the row to the current Position
         when NumVal = 1;
           ClearData(CurrRow:1:CurrRow:CurrCol);

       //  Clear the entire row
         when NumVal = 2;
           ClearData(CurrRow:1:CurrRow:80);

         endsl;

       endsr;
       // *********************************************************************

       //  Set screen attributes

       begsr SetAttr;

         for c = 1 to 4;
           tAttr = %subst(AttrStr:c:1);
           select;
           when tAttr = '0';
             AttrControl = *blanks;
           when tAttr = '1';
             White = 'W';
           when tAttr = '4';
             Underline = 'U';
             UnderlineRcvd = *on;
           when tAttr = '5';
             Blink = 'B';
           when tAttr = '7';
             ReverseImage = 'R';
           endsl;
         endfor;

         select;
       //  normal
         when AttrControl = *blanks;
           CurrAttr = NormalAttr;
       //  Reverse image
         when AttrControl = '  R ';
           CurrAttr = x'21';
       //  White (hi intensity)
         when AttrControl = 'W   ';
           CurrAttr = x'22';
       //  White, reverse image
         when AttrControl = 'W R ';
           CurrAttr = x'23';
       //  Underline
         when AttrControl = ' U  ';
           CurrAttr = x'24';
       //  Underline, reverse image
         when AttrControl = ' UR ';
           CurrAttr = x'25';
       //  White, underline
         when AttrControl = 'WU  ';
           CurrAttr = x'26';
       //  White, underline, reverse image (non-display)
         when AttrControl = 'WUR ';
           CurrAttr = x'27';
       //  Blink
         when AttrControl = '   B';
           CurrAttr = x'2A';
       //  Reverse image, blink
         when AttrControl = '  RB';
           CurrAttr = x'2B';
       //  Reverse image, underline, blink
         when AttrControl = ' URB';
           CurrAttr = x'2D';
       //  Underline, blink
         when AttrControl = ' U B';
           CurrAttr = x'2E';
       //  Ignore anything else
         other;
           CurrAttr = NormalAttr;
         endsl;

       endsr;

       // *********************************************************************

       //  Parse wait string array

       begsr ParseWaitStr;

         WaitCnt = 0;
         WaitOpt = *all' ';

         for w1 = 1 to 20;

       //  Blank element indicates end
           if WaitArr(w1) = *blanks;
             leave;
           endif;

       //  Bypass invalid row/column/key
           WaitCtl = WaitArr(w1);
           if WaitRow < 1 or WaitRow > 24;
             iter;
           endif;
           if WaitCol < 1 or WaitCol > 80;
             iter;
           endif;
           if WaitStr = *blanks;
             iter;
           endif;

       //  Set Position and key length
           WaitCnt = w1;
           WaitPos(Waitcnt) = ((Waitrow-1)*80)+Waitcol;
           w2 = %len(%trim(WaitStr));
           WaitKey(Waitcnt) = Waitstr;

       //  If the last character is %CR it indicates trailing blanks
           if W2 > 3
              and %subst(WaitStr:w2-2:3) = '%CR';
             %subst(WaitKey(WaitCnt):w2-2) = *blanks;
             w2 = w2 - 3;
           endif;

           W3 = WaitPos(WaitCnt) + w2;
           if w3 > 1921;
             w2 = 1921 - WaitPos(WaitCnt);
           endif;
           WaitLen(WaitCnt) = w2;

       //  Default to 'OR' (start of sequence)
           if WaitCnt = 1 or WaitOption = 'O' or WaitOption = 'o'
              or WaitOption = *blanks;
             WaitOpt(WaitCnt) = 'O';
           else;
             WaitOpt(WaitCnt) = 'A';
           endif;

         endfor;

       endsr;

       // *********************************************************************

       //  Evaluate wait parameters

       begsr CheckWaitStr;

         ParmWaitPtr = *zeros;

         for w1 = 1 to WaitCnt;

       //  Ignore invalid keys
           if WaitOpt(w1) = *blanks;
             iter;
           endif;

       //  Start of new 'OR' sequence
           if WaitOpt(w1) = 'O';

       //  All keys have matched in this sequence. We are done
             if ParmWaitPtr <> 0;
               leave;
             endif;

       //  Reset pointer
             ParmWaitptr = w1;

       //  'AND' entry. Only continue if we have already matched.
           else;
             if ParmWaitptr = 0;
               iter;
             endif;

           endif;

       //  Compare the wait key with the screen buffer
           w2 = WaitPos(w1);
           w3 = WaitLen(w1);
           clear WaitCompStr;
           WaitCompStr = %subst(ScreenChar:w2:w3);

       //  No match. Zero out pointer.
           if WaitCompStr <> WaitKey(w1);
             ParmWaitptr = 0;
           endif;

         endfor;

       endsr;

       // *********************************************************************

       //  Connect to the telnet server on the iSeries

       begsr ConnectToiSeries;

       //  Set logging parameters and open logging file
         if %open(TNAPILOG);
           close TNAPILOG;
         endif;

         if %open(TNAPIBUFF);
           close TNAPIBUFF;
         endif;

         monitor;
           open TNAPIBUFF;
           LogBuffer = *on;
         on-error;
           LogBuffer = *off;
         endmon;

       //  If log file name specified, override to that file
         Logging = *off;
         select;
         when  ParmLogFile = '*NONE';
         when %subst(ParmLogFile:1:1) = '*';
           select;
           when ParmLogFile = '*JOBNAME';
             JobMbr = SDSJobName;
           when ParmLogFile = '*USER';
             JobMbr = SDSUser;
           when ParmLogFile = '*JOBNBR';
             JobMbr = 'JOB' + SDSJob#;
           endsl;
           Cmd = 'ADDPFM FILE(TNAPILOG) MBR(' + JobMbr + ')';
           Exsr RunCmd;
           Cmd = 'OVRDBF FILE(TNAPILOG) TOFILE(TNAPILOG) ' +
            'MBR(' + JobMbr + ') LVLCHK(*NO)';
           Exsr RunCmd;
           Logging = *on;
         when ParmLogFile <> 'TNAPILOG'
            and ParmLogFile <> *blanks;
           Cmd = 'OVRDBF FILE(TNAPILOG) TOFILE(' +
            %trim(ParmLogFile) + ') LVLCHK(*NO)';
           exsr RunCmd;
           if not CmdError;
             Logging = *on;
           endif;
         endsl;

       //  Try to open the file. If error, turn logging off
         if Logging;
           monitor;
             open TNAPILOG;
             Logging = *on;
             LogJob = SDSJob# + '/' + %trim(SDSUser) + '/' + SDSJobName;
           on-error;
             Logging = *off;
           endmon;

         endif;

         if %open(TNAPIDSPF);
           close TNAPIDSPF;
         endif;

       //  try to open TNAPIDSPF. If error occurs, turn off display
         DisplayScreen = *off;
         if ParmDisplay = 'Y';
           monitor;
             open TNAPIDSPF;
             DisplayScreen = *on;
           on-error;
           endmon;
         endif;

       //  Set up special strings
         NormalDisplay = ESCChar + LeftBracket + '0m';
         SendEOT = ESCChar + LeftBracket + 'A'
          + ESCChar + LeftBracket + 'B';
         RecvEOT = ESCChar + LeftBracket + '1A'
          + ESCChar + LeftBracket + '1B';
         MoveToTop = ESCChar + LeftBracket + '1;';
         MoveToBtm = ESCChar + LeftBracket + '24;';
         BinaryValue = 39;
         %subst(PosCsrStr:1:1) = Char;
         BinaryValue = 74;
         %subst(PosCsrStr:2:1) = Char;

       //  Previous call didn't end properly
         if SocketID >= 0;
           rc = close(SocketID);
         endif;

       //  Set socket attributes
         SocketID = Socket(AF_INET:SOCK_STREAM:UNUSED);
         if SocketID < 0;
           exsr Eoj;
         endif;

       //  Connect the socket to the telnet port on the desired system
         Svrfamily = AF_INET;
         Svrport  = 23;
         LogSystem = ParmSystem;

       //  If system name is blank connect to loopback
         if Parmsystem = *blanks;
           SvrAddr = GetHostId;
         else;
       //  See if an IP address was passed in
           SvrAddr = inet_addr(%trim(ParmSystem));
           if SvrAddr = INADDR_NONE;
       //  Must be host name
             p_hostent = gethostbyname(%trim(ParmSystem));
             if p_hostent = *NULL;
               exsr EOJ;
             endif;
             SvrAddr = h_addr;
           endif;
         endif;

       // Connect to socket
         SvrNull   = X'0000000000000000';
         SVRADDRLEN= %Size(SvrAddress);
         rc = Connect(SocketID:%addr(SvrAddress):SVRADDRLEN);

       // Unable to connect
         if rc < 0;
           exsr Eoj;
         endif;

       //  Set the socket for non-blocking I/O
         rc = ioctl(SocketID:FIONBIO:NONBLOCKIO);
         if rc < 0;
           exsr Eoj;
         endif;

        // Create a file desciption set for this socket
         SaveSet = *allx'00';
         fd_set(SocketID:SaveSet);

         CurrRow = 1;
         CurrCol = 1;
         clear ScreenChar;
         clear ScreenAttr;
         UnderlineRcvd = *off;

       //  Send string to initiate iSeries VT100 conversation:
       //  X, backspace, space, backspace
         SendBuff = 'X' +
          ESCChar + LeftBracket + 'D' + ' ' +
          ESCChar + LeftBracket + 'D';
         SendBuff = %trimr(SendBuff) + SendEOT;
         exsr WriteData;

       //  Wait for the signon screen
         Timeout = ThirtySeconds;
         exsr RecvData;

         if ParmReturn = '2';
           *in25 = *on;
           exsr LogData;
           exsr EOJ;
         endif;

         if Logging;
           exsr LogScreen;
         endif;
         LastSystem = ParmSystem;

         FirstTime = *off;

       endsr;

       // *********************************************************************

       //  Log function key

       begsr LogFctKey;

         LogField = '<' + %trim(ParmFctKey) + '>';
         *in21 = *on;
         exsr Logdata;

       endsr;

       // *********************************************************************

       //  Log screen data to file
       begsr LogScreen;

         Logrow = CurrRow;
         LogCol = CurrCol;

         *in22 = *on;

       // Screen does not contain any underlined characters.
         if not UnderlineRcvd;

           for r = 1 to 24;
             LogScreenBuff = ScreenRow(r);
             *in23 = *on;
             exsr LogData;
           endfor;

         else;

       // Convert underlined blanks to underscores
           p = 1;
           for r = 1 to 24;
             for c = 1 to 80;
               Tchar = ScreenCol(p);
               if Tchar = *blank;
                 TAttr = AttrCol(p);
                 if TAttr > x'23' and TAttr < x'27';
                   Tchar = '_';
                 endif;
               endif;
               LogScreenChar(c) = Tchar;
               p = p+1;
             endfor;
             *in23 = *on;
             exsr LogData;
           endfor;

         endif;

         LogScreenBuff = Seperator;
         *in23 = *on;
         exsr LogData;

       endsr;

       // *********************************************************************

       //  Move cursor to desired location on screen by sending
       //   cursor control codes

       begsr MoveCursor;

         PosRow = 0;
         PosCol = 0;
         clear CsrLocStr;

       // Validate row and column
         monitor;
           PosRow = %int(PosRow$);
           PosCol = %int(PosCol$);
         on-error;
         endmon;

         if PosRow > 0 and PosCol > 0
            and PosRow < 25 and PosCol < 81;

       // Determine if the cursor needs to be moved up or down
           PosCtl = 'A';
           PosCount = CurrRow - PosRow;
           if PosCount < 0;
             PosCount = PosCount * -1;
             PosCtl = 'B';
           endif;

       // Move cursor up or down
           for p = 1 to PosCount;
             CsrLocStr = %trim(CsrLocStr) + PosCsrStr + PosCtl;
           endfor;

       // Determine if cursor needs to be moved left or right
           PosCtl = 'D';
           PosCount = CurrCol - PosCol;
           if PosCount < 0;
             PosCount = PosCount * -1;
             PosCtl = 'C';
           endif;

       // Move cursor left or right
           for p = 1 to PosCount;
             CsrLocStr = %trim(CsrLocStr) + PosCsrStr + PosCtl;
           endfor;
         endif;

       endsr;

       // *********************************************************************

       //  Run command
       begsr RunCmd;

         monitor;
           ExecCmd(Cmd:CMdLen);
           CmdError = *off;
         on-error;
           CmdError = *on;
         endmon;

       endsr;

       // *********************************************************************

       //  Log raw data to file
       begsr LogData;

         CurrTime = %timestamp();
         DateTime = %subst(TimeStamp:1);
         except;
         feod TNAPILOG;
         *in21 = *off;
         *in22 = *off;
         *in23 = *off;
         *in24 = *off;
         *in25 = *off;

       endsr;

       // *********************************************************************

       //  End of job
       begsr Eoj;

         ParmReturn = '3';
         rc = close(SocketID);
         *INLR = *on;
         return;

       endsr;

       // ********************************************************************

      /end-free

     OTNAPILOG  e    21
     O                       datetime
     O                                           29 'SENT:'
     O                       LogField            80
     O          e    22
     O                       datetime
     O                                           30 'RECV: '
     O                       LogRow        x
     O                                              ','
     O                       LogCol        x
     O                       parmreturn          39
     O                       LogJob              68
     O                       LogSystem           79
     O          e    23
     O                       LogScreenBuff
     O          e    25
     O                       datetime
     O                                        +   2 'Unable to connect to:'
     O          e    25
     O                       parmsystem
     O          e    26
     O                       LogSeperator
     O          e    26
     O                                              'Log data'
     O                       datetime         +   2
     O          e    26
     O                       Logdta
     O          e    26
     O                       LogSeperator

     OTNAPIBUFF e    27
     O                       datetime
     O                                              '<'
     O                       readbuff         +   1
     O          e    28
     O                       datetime
     O                                              '>'
     O                       SendBuff         +   1
       // *********************************************************************

       //  Clear data from parts of the screen
     P ClearData       B

     D ClearData       PI
     D   StartRow                     2  0 const
     D   StartCol                     2  0 const
     D   EndRow                       2  0 const
     D   EndCol                       2  0 const

     D len             S             10I 0
     D start           S             10I 0

      /free
       start = ((StartRow-1)*80) + StartCol;
       Len  = ((EndRow-1)*80) + EndCol - start + 1;
       %subst(ScreenChar:start:len) = *blanks;
       %subst(ScreenAttr:start:len) = *blanks;
      /end-free
     P                 E

       // *********************************************************************

       //  Move character to send buffer
     P MoveCharToBuff  B

      /free
       SendLen = SendLen + 1;
       SendChar(SendLen) = char;
      /end-free
     P                 E
       // ********************************************************************

     P CheckForData    B                   export
     D CheckForData    PI              n
     D  SocketCheck                  10i 0
     D  TimeOut                      15p 0
       //  Check for data

     D select          pr            10i 0 extproc('select')
     D  nfds                         10i 0 value
     D  readfds                        *   value
     D  writefds                       *   value
     D  exceptfds                      *   value
     D  timeout                        *   value

     D rc              s             10i 0
     D CheckID         s             10i 0

      /free
       ReadSet = SaveSet;
       CheckID = SocketCheck + 1;
       rc=select(CheckID:%addr(readset):*null:*null:
        %addr(timeval));
       if readset = saveset;
         return *on;
       else;
         return *off;
       endif;

      /end-free
     P                 E

     P*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      * SK
     P* Set a File Descriptor in a set ON...  for use w/Select()
     P*
     P*      peFD = descriptor to set on
     P*      peFDSet = descriptor set
     P*
      *
     P FD_SET          B                   EXPORT
     D FD_SET          PI
     D   peFD                        10I 0
     D   peFDSet                     28A
     D wkByteNo        S              5I 0
     D wkMask          S              1A
     D wkByte          S              1A
     C                   callp     CalcBitPos(peFD:wkByteNo:wkMask)
     c                   eval      wkByte = %subst(peFDSet:wkByteNo:1)
     c                   biton     wkMask        wkByte
     c                   eval      %subst(peFDSet:wkByteNo:1) = wkByte
     P                 E

     P*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      * SK
     p*  This is used by the FD_SET/FD_CLR/FD_ISSET procedures to
     p*  determine which byte in the 28-char string to check,
     p*  and a bitmask to check the individual bit...
     p*
     p*  peDescr = descriptor to check in the set.
     p*  peByteNo = byte number (returned)
     p*  peBitMask = bitmask to set on/off or test
     P*
      *
     P CalcBitPos      B
     D CalcBitPos      PI
     D    peDescr                    10I 0
     D    peByteNo                    5I 0
     D    peBitMask                   1A
     D dsMakeMask      DS
     D   dsZeroByte            1      1A
     D   dsMask                2      2A
     D   dsBitMult             1      2U 0 INZ(0)
     C     peDescr       div       32            wkGroup           5 0
     C                   mvr                     wkByteNo          2 0
     C                   div       8             wkByteNo          2 0
     C                   mvr                     wkBitNo           2 0
     C                   eval      wkByteNo = 4 - wkByteNo
     c                   eval      peByteNo = (wkGroup * 4) + wkByteNo
     c                   eval      dsBitMult = 2 ** wkBitNo
     c                   eval      dsZeroByte = x'00'
     c                   eval      peBitMask = dsMask
     P                 E

** FKNAME Function key name table
ENTER
F1
F2
F3
F4
F5
F6
F7
F8
F9
F01
F02
F03
F04
F05
F06
F07
F08
F09
F10
F11
F12
F13
F14
F15
F16
F17
F18
F19
F20
F21
F22
F23
F24
PAGEDN
PAGEUP
ATTN
SYSREQ
HOME
PRINT
RESET
REDRAW
HELP
** FKSEQ Function key mapping codes
13
39241
39242
39243
39244
39245
39246
39247
39248
39249
39241
39242
39243
39244
39245
39246
39247
39248
39249
39240
39096
39126
39079
39124
39123
39091
39108
39186
39080
39092
39077
39093
39109
39078
55
02
01
03
15
37
18
39211
39200
** CCNAME Cursor control names
BS Backspace
DN Down
LF Left
RT Right
UP Up
DL Delete
DP Dup
ER Erase input
BT Back tab
FE Field exit
FP Field plus
FM Field minus
IN Insert
NL New line
TB Tab
** CCSEQ Cursor control codes
2200
3974B
3974D
3974C
3974A
0700
3900D
4500
3905
3900X
3900X
3900M
3900I
3937
0500
